
import java.io.*;
import java.util.*;

import javax.vecmath.*;

import org.autochthonous.graphics.rt.*;
import org.autochthonous.graphics.rt.tools.io.*;
import org.autochthonous.graphics.rt.util.ParametersMap;
import org.autochthonous.graphics.rt.shaders.ct.*;

/**
 * A tiling bump map shader.
 * <p>
 * Example:
 * <code>
 * <pre>
 *      &lt;BumpMappedMetal diffuse="0.9" specular="0.1"&gt;
 *         &lt;Tile image="tile.pix" repeating="true"&gt;
 *      &lt;/BumpMappedMetal&gt;
 * </pre>
 * </code>
 * <p>
 * Document Type Definition:
 * <code>
 * <pre>
 *     &lt;!ENTITY %Float    "CDATA"&gt;
 *     &lt;!ENTITY %Color3f  "CDATA"&gt;
 *     &lt;!ENTITY %Vector2f "CDATA"&gt;
 *
 *     &lt;!ELEMENT BumpMappedMetal (Tile)&gt;
 *     &lt;!ATTLIST BumpMappedMetal
 *         ambient              (%Color3f; | %Float;) "0, 0, 0"
 *         specular             (%Color3f; | %Float;) "0, 0, 0"
 *         rotate               %Float; "0"
 *         scale                %Vector2f; "1, 1"&gt;
 * </pre>
 * </code>
 */
public class BumpMappedMetal
    extends Metal
{
    Tile        tile;
    float       sScale = 1;
    float       tScale = 1;
    float       cosAng = 1;
    float       sinAng = 0;
    Vector3f    pturb = new Vector3f();
    float[]     deltas = new float[2];

    public BumpMappedMetal(
        World   world,
        Tile    tile)
    {
        super(world);

        this.tile = tile;
    }

    private BumpMappedMetal(
        World   world)
    {
        super(world);
    }

    public static Shader getInstance(
        World         world,
        RMLReader       in)
        throws IOException
    {
        BumpMappedMetal   m = new BumpMappedMetal(world);

        m.readObject(in);

        return m;
    }

    protected void processObject(
        Object  obj)
    {
        if (obj instanceof Tile)
        {
            this.tile = (Tile)obj;
        }
        else
        {
            super.processObject(obj);
        }
    }

    /**
     * handle the parameters we want, passing the extras
     * to our super class for processing.
     */
    protected void processParameters(
        ParametersMap    params)
    {
        super.processParameters(params);

        if (params.get("scale") != null)
        {
            StringTokenizer st = new StringTokenizer((String)params.get("scale"), " ,");
            setScale(Float.parseFloat(st.nextToken()), Float.parseFloat(st.nextToken()));
        }

        setRotation(params.getFloat("rotate", "0"));
    }

    /**
     * set a rotation in degrees (around "u") for this tiling.
     */
    public void setRotation(
        float   angle)
    {
        this.cosAng = (float)Math.cos(angle * Math.PI / 180.0);
        this.sinAng = (float)Math.sin(angle * Math.PI / 180.0);
    }

    /**
     * set the (s, t) scaling for this tiling.
     */
    public void setScale(
        float    sScale,
        float    tScale)
    {
        this.sScale = sScale;
        this.tScale = tScale;
    }

    /**
     * the shade routine. If you want more than the below, this is
     * the method to override.
     */
    public void shade(
        Hit        hit,
        Pixel      p)
    {
        p.set(0, 0, 0, 1);

        float       s = hit.S() / sScale;
        float       t = hit.T() / tScale;

        if (tile.getDeltas(s * cosAng - t * sinAng, t * cosAng + s * sinAng, deltas))
        {
            hit.perturbN(deltas[0] * 6, deltas[1] * 6);
        }

        LightDataList   contribs = hit.visibleLights(world);

        p.scaledAdd(hit.color(), ambientContribution(world));

        p.scaledAdd(hit.color(), specular,
                specularContribution(contribs, hit, roughness));

        p.scale(alpha);
    }
}
