package org.autochthonous.graphics.rt.models;

import java.io.*;
import java.util.*;

import javax.vecmath.*;

import org.autochthonous.graphics.rt.*;
import org.autochthonous.graphics.rt.util.*;
import org.autochthonous.graphics.rt.objects.*;
import org.autochthonous.graphics.rt.tools.io.*;

/**
 * a parser for the DXF files.
 * <pre>
 * &lt;DXFFile geometry="ant.dxf"&gt;
 *     &lt;Plastic name="body" diffuse="1.0, 0.0, 0.0"/&gt;
 * &lt;/DXFFile&gt;
 * </pre>
 * <p>
 * Shaders specified within the body of the file object are used for 
 * objects appearing in the DXF layer with the same name.
 */
public class DXFFile
    extends Aggregate
{
    private HashMap             shaderTable = new HashMap();
    private String              geometry;

    static private class DXFFileReader
    {
        private static final int    NUM_FLAGS       = 6;
        private static final int    BIT_FLAG        = 0;

        private static final int    M               = 1;
        private static final int    N               = 2;

        private static final int    CLOSED          = 1;
        private static final int    THREEDPOLYLINE  = 8;
        private static final int    THREEDPOLYMESH  = 16;
        private static final int    POLYFACEMESH    = 64;

        private class DXFVertex
        {
            Point3f point = new Point3f();
            float   extLen = 0;
            int[]   flags = new int[NUM_FLAGS];
        }

        private BufferedReader  in;
        private HashMap         shaderTable;
        private HashMap         blocks = new HashMap();

        DXFFileReader(
            BufferedReader  in,
            HashMap         shaderTable)
        {
            this.in = in;
            this.shaderTable = shaderTable;
        }

        void readBlock()
            throws IOException
        {
            Aggregate   a = new Aggregate();
            int[]       flags = new int[NUM_FLAGS];
            String      blockName = null;
            String      line;

            while ((line = in.readLine()) != null)
            {
                int         code = Integer.parseInt(line.trim());

                if (code == 0)
                {
                    while ((line = in.readLine()) != null)
                    {
                        if (line.equals("3DFACE"))
                        {
                            read3DFace(a);
                        }
                        else if (line.equals("CIRCLE"))
                        {
                            readCircle(a);
                        }
                        else if (line.equals("POLYLINE"))
                        {
                            readPolyline(a);
                        }
                        else if (line.equals("LINE"))
                        {
                            readLine(a);
                        }
                        else if (line.equals("SOLID"))
                        {
                            readSolid(a);
                        }
                        else if (line.equals("INSERT"))
                        {
                            readInsert(a);
                        }
                        else if (line.equals("ENDBLK"))
                        {
                            if (blockName != null)
                            {
                                blocks.put(blockName, a);
                            }
                            return;
                        }
                        else
                        {
                            skipEntity();
                        }
                    }
                }
                else
                {
                    line = in.readLine();

                    switch (code)
                    {
                    case 2:
                        blockName = line;
                        break;
                    case 1:
                    case 3:
                    case 5:
                    case 8:
                    case 10:
                    case 20:
                    case 30:
                        break;
                    case 70:
                    case 71:
                    case 72:
                    case 73:
                    case 74:
                    case 75:
                        flags[code - 70] = Integer.parseInt(line.trim());
                        break;
                    default:
                        System.out.println("unknown code encountered in block: " + code);
                    }
                }
            }
        }

        void readBlocks()
            throws IOException
        {
            String  line = in.readLine();

            while ((line = in.readLine()) != null)
            {
                if (line.equals("BLOCK"))
                {
                    readBlock();
                }
                else if (line.equals("ENDSEC"))
                {
                    return;
                }
            }
        }

        void readCircle(
            Aggregate   a)
            throws IOException
        {
            Shader      shader = null;
            String      line;
            float       extLen, radius;
            float       x, y, z;
            float       xDir, yDir, zDir;
            Point3f     last = null;
            int[]       flags = new int[NUM_FLAGS];

            extLen = radius = 0;
            x = y = z = 0;

            xDir = yDir = 0;
            zDir = 1;

            while ((line = in.readLine()) != null)
            {
                int code = Integer.parseInt(line.trim());

                if (code == 0)
                {
                    if (extLen != 0.0)
                    {
                        Cylinder c = new Cylinder(
                                new Point3f(x, y, z),
                                new Point3f(x + extLen * xDir,
                                            y + extLen * yDir,
                                            z + extLen * zDir),
                                radius);

                        c.setShader(shader);

                        a.add(c);
                    }
                    return;
                }

                line = in.readLine();

                switch (code)
                {
                case 8:         // layer
                    shader = (Shader)shaderTable.get(line);
                    break;
                case 62:
                    // color
                    break;
                case 5:
                case 67:
                    // handle
                    break;
                case 39:
                    extLen = Float.parseFloat(line);
                    break;
                case 10:
                    x = Float.parseFloat(line);
                    break;
                case 20:
                    y = Float.parseFloat(line);
                    break;
                case 30:
                    z = Float.parseFloat(line);
                    break;
                case 40:
                    radius = Float.parseFloat(line);
                    break;
                case 70:
                case 71:
                case 72:
                case 73:
                case 74:
                case 75:
                    flags[code - 70] = Integer.parseInt(line.trim());
                    break;
                case 210:                 // EXTRU_X
                    xDir = Float.parseFloat(line);
                    break;
                case 220:                 // EXTRU_Y
                    yDir = Float.parseFloat(line);
                    break;
                case 230:                 // EXTRU_Z
                    zDir = Float.parseFloat(line);
                    break;
                default:
                    System.out.println("unknown code encountered in Circle: " + code);
                }
            }

            throw new IOException("malformed Circle");
        }
        
        private DXFVertex readVertex()
            throws IOException
        {
            DXFVertex   v = new DXFVertex();
            String      line;

            while ((line = in.readLine()) != null)
            {
                int code = Integer.parseInt(line.trim());
                if (code == 0)
                {
                    return v;
                }
                
                line = in.readLine();

                switch (code) {
                case 5:
                case 8:         // layer
                    break;
                case 6:
                case 62:
                    // color
                    break;
                case 67:
                    // handle
                    break;
                case 10:
                case 11:
                case 12:
                case 13:
                case 14:
                case 15:
                case 16:
                    v.point.x = Float.parseFloat(line);
                    break;
                case 20:
                case 21:
                case 22:
                case 23:
                case 24:
                case 25:
                case 26:
                    v.point.y = Float.parseFloat(line);
                    break;
                case 30:
                case 31:
                case 32:
                case 33:
                case 34:
                case 35:
                case 36:
                    v.point.z = Float.parseFloat(line);
                    break;
                case 39:
                    v.extLen = Float.parseFloat(line);
                    break;
                case 41:
                case 42:
                case 43:
                case 50:
                    // ignore
                    break;
                case 70:
                case 71:
                case 72:
                case 73:
                case 74:
                case 75:
                    v.flags[code - 70] = Integer.parseInt(line.trim());
                    break;
                default:
                    System.out.println("unknown code encountered in vertex: " + code);
                }
            }

            return null;
        }

        private void addPolygon(
            Aggregate   a,
            Point3f[]   pBuf,
            Point3f     v1,
            Point3f     v2,
            float       extLen,
            Vector3f    dir,
            Shader      shader)
        {
            pBuf[0] = v1;
            pBuf[1] = v2;

            pBuf[2] = new Point3f(
                        pBuf[1].x + extLen * dir.x,
                        pBuf[1].y + extLen * dir.y,
                        pBuf[1].z + extLen * dir.z);

            pBuf[3] = new Point3f(
                        pBuf[0].x + extLen * dir.x,
                        pBuf[0].y + extLen * dir.y,
                        pBuf[0].z + extLen * dir.z);

            try
            {
                Polygon p = new Polygon(pBuf);

                p.setShader(shader);

                a.add(p);
            }
            catch (IllegalArgumentException e)
            {
                // a dud ignore
            }
        }

        /**
         * read in a solid
         */
        void readSolid(
            Aggregate   a)
            throws IOException
        {
            Shader      shader = null;
            ArrayList   verts = new ArrayList();
            String      line;
            float       x, y, z;
            Point3f     last = null;
            boolean     extrusion = false;
            Vector3f    dir = new Vector3f(0, 0, 1);
            float       extLen = 0;
            int[]       flags = new int[NUM_FLAGS];

            x = y = z = 0;

            while ((line = in.readLine()) != null)
            {
                int code = Integer.parseInt(line.trim());
                if (code == 0)
                {
                    //
                    // vertices are originally 1, 3, 4, 2 so if we have 4
                    // we need to reorder them to ease processing.
                    //

                    if (verts.size() == 4)
                    {
                        Object  o = verts.get(3);

                        verts.set(3, verts.get(2));
                        verts.set(2, verts.get(1));
                        verts.set(1, o);
                    }

                    Point3f[] points = (Point3f[])verts.toArray(new Point3f[verts.size()]);

                    Polygon p = new Polygon(points);

                    p.setShader(shader);

                    a.add(p);

                    if (extrusion)
                    {
                        Point3f[] vs = new Point3f[4];

                        for (int i = 0; i != points.length - 1; i++)
                        {
                            addPolygon(a, vs, points[i], points[i + 1], extLen, dir, shader); 
                        }

                        addPolygon(a, vs, points[points.length - 1], points[0], extLen, dir, shader); 

                        for (int i = 0; i != points.length; i++)
                        {
                            points[i].x += extLen * dir.x;
                            points[i].y += extLen * dir.y;
                            points[i].z += extLen * dir.z;
                        }

                        p = new Polygon(points);

                        p.setShader(shader);

                        a.add(p);
                    }
                    return;
                }

                line = in.readLine();

                switch (code)
                {
                case 8:         // layer
                    shader = (Shader)shaderTable.get(line);
                    break;
                case 62:
                    // color
                    break;
                case 5:
                    break;
                case 10:
                case 11:
                case 12:
                case 13:
                case 14:
                case 15:
                case 16:
                    x = Float.parseFloat(line);
                    break;
                case 20:
                case 21:
                case 22:
                case 23:
                case 24:
                case 25:
                case 26:
                    y = Float.parseFloat(line);
                    break;
                case 30:
                case 31:
                case 32:
                case 33:
                case 34:
                case 35:
                case 36:
                    z = Float.parseFloat(line);
                    //
                    // if it's a triangle the last point gets repeated...
                    //
                    if (last != null)
                    {
                        if (last.x != x || last.y != y || last.z != z)
                        {
                            last = new Point3f(x, y, z);
                            verts.add(last);
                        }
                    }
                    else
                    {
                        last = new Point3f(x, y, z);
                        verts.add(last);
                    }
                    break;
                case 39:
                    extrusion = true;
                    extLen = Float.parseFloat(line);
                    break;
                case 70:
                case 71:
                case 72:
                case 73:
                case 74:
                case 75:
                    flags[code - 70] = Integer.parseInt(line.trim());
                    break;
                case 210:                 // EXTRU_X
                    extrusion = true;
                    dir.x = Float.parseFloat(line);
                    break;
                case 220:                 // EXTRU_Y
                    extrusion = true;
                    dir.y = Float.parseFloat(line);
                    break;
                case 230:                 // EXTRU_Z
                    extrusion = true;
                    dir.z = Float.parseFloat(line);
                    break;
                default:
                    if (code < 1000)
                    {
                        System.out.println("unknown code encountered in solid: " + code);
                    }
                }
            }

            throw new IOException("malformed solid");
        }

        /**
         * read in an insert
         */
        void readInsert(
            Aggregate   a)
            throws IOException
        {
            Shader      shader = null;
            ArrayList   verts = new ArrayList();
            String      line;
            float       x, y, z;
            float       xScale, yScale, zScale;
            Point3f     last = null;
            boolean     extrusion = false;
            Vector3f    dir = new Vector3f(0, 0, 1);
            float       extLen = 0;
            int[]       flags = new int[NUM_FLAGS];
            String      blockName = null;
            float       rotAngle = 0;
            float       endAngle = 0;

            x = y = z = 0;
            xScale = yScale = zScale = 1;

            while ((line = in.readLine().trim()) != null)
            {
                int code = Integer.parseInt(line);
                if (code == 0)
                {
                    if (blockName != null)
                    {
                        Aggregate   blk = (Aggregate)blocks.get(blockName);

                        if (blk != null)
                        {
                            Aggregate   newAg = new Aggregate();
                            Transform   trans = new Transform();

                            trans.rotate(0, 0, 1, rotAngle);
                            trans.scale(xScale, yScale, zScale);
                            trans.translate(x, y, z);

                            TraceableIterator    it = blk.traceables(); 

                            while (it.hasNext())
                            {
                                try
                                {
                                    Renderable  o = (Renderable)it.next().clone();
                                    o.transform(trans);

                                    newAg.add(o);
                                }
                                catch (CloneNotSupportedException e)
                                {
                                    System.err.println(e);
                                }
                            }

                            a.add(newAg);
                        }
                    }

                    return;
                }

                line = in.readLine().trim();

                switch (code)
                {
                case 2:         // block name
                    blockName = line;
                    break;
                case 8:         // layer
                    shader = (Shader)shaderTable.get(line);
                    break;
                case 6:
                case 62:
                    // color
                    break;
                case 5:
                case 67:
                    // handle
                    break;
                case 10:
                    x = Float.parseFloat(line);
                    break;
                case 20:
                    y = Float.parseFloat(line);
                    break;
                case 30:
                    z = Float.parseFloat(line);
                    break;
                case 41:
                    xScale = Float.parseFloat(line);
                    break;
                case 42:
                    yScale = Float.parseFloat(line);
                    break;
                case 43:
                    zScale = Float.parseFloat(line);
                    break;
                case 50:
                    rotAngle = Float.parseFloat(line);
                    break;
                case 51:
                    endAngle = Float.parseFloat(line);
                    break;
                case 70:
                case 71:
                case 72:
                case 73:
                case 74:
                case 75:
                    flags[code - 70] = Integer.parseInt(line);
                    break;
                default:
                    if (code < 1000)
                    {
                        System.out.println("unknown code encountered in insert: " + code);
                    }
                }
            }

            throw new IOException("malformed insert");
        }

        /**
         * read in a line
         */
        void readLine(
            Aggregate   a)
            throws IOException
        {
            Shader      shader = null;
            ArrayList   verts = new ArrayList();
            String      line;
            float       x, y, z;
            float       xDir, yDir, zDir;
            boolean     extrusion = false;
            int[]       flags = new int[NUM_FLAGS];
            Vector3f    dir = new Vector3f(0, 0, 1);
            float       extLen = 0;
            Point3f     start = new Point3f();
            Point3f     end = new Point3f();

            while ((line = in.readLine()) != null)
            {
                int code = Integer.parseInt(line.trim());
                if (code == 0)
                {
                    if (extrusion)
                    {
                        Point3f[]     vs = new Point3f[4];
                        
                        if (!start.equals(end))
                        {
                            addPolygon(a, vs, start, end, extLen, dir, shader);
                        }

                        return;
                    }

                    return;     // just a line
                }

                line = in.readLine();

                switch (code)
                {
                case 8:         // layer
                    shader = (Shader)shaderTable.get(line);
                    break;
                case 6:
                case 62:
                    // color
                    break;
                case 5:
                case 67:
                    // handle
                    break;
                case 10:
                    start.x = Float.parseFloat(line);
                    break;
                case 20:
                    start.y = Float.parseFloat(line);
                    break;
                case 30:
                    start.z = Float.parseFloat(line);
                    break;
                case 11:
                    end.x = Float.parseFloat(line);
                    break;
                case 21:
                    end.y = Float.parseFloat(line);
                    break;
                case 31:
                    end.z = Float.parseFloat(line);
                    break;
                case 39:
                    extrusion = true;
                    extLen = Float.parseFloat(line);
                    break;
                case 70:
                case 71:
                case 72:
                case 73:
                case 74:
                case 75:
                    flags[code - 70] = Integer.parseInt(line.trim());
                    break;
                case 210:                 // EXTRU_X
                    extrusion = true;
                    dir.x = Float.parseFloat(line);
                    break;
                case 220:                 // EXTRU_Y
                    extrusion = true;
                    dir.y = Float.parseFloat(line);
                    break;
                case 230:                 // EXTRU_Z
                    extrusion = true;
                    dir.z = Float.parseFloat(line);
                    break;
                default:
                    if (code < 1000)
                    {
                        System.out.println("unknown code encountered in line: " + code);
                    }
                }
            }

            throw new IOException("malformed line");
        }

        private void addTriangle(
            Aggregate   a,
            Shader      shader,
            Point3f[]   tri)
        {
            try
            {
                Polygon p = new Polygon(tri);

                p.setShader(shader);

                a.add(p);
            }
            catch (IllegalArgumentException e)
            {
                // ignore
            }
        }

        /**
         * read in a polyline
         */
        void readPolyline(
            Aggregate   a)
            throws IOException
        {
            Shader      shader = null;
            ArrayList   verts = new ArrayList();
            String      line;
            float       x, y, z;
            float       xDir, yDir, zDir;
            boolean     extrusion = false;
            int[]       flags = new int[NUM_FLAGS];
            Vector3f    dir = new Vector3f(0, 0, 1);
            float       extLen = 0;
            DXFVertex   start = new DXFVertex();

            while ((line = in.readLine()) != null)
            {
                int code = Integer.parseInt(line.trim());

                if (code == 0)
                {
                    line = in.readLine();

                    if (line.equals("VERTEX"))
                    {
                        while (line.equals("VERTEX"))
                        {
                            DXFVertex   v = readVertex();

                            verts.add(v);
                            line = in.readLine();
                        }

                        if (line.equals("SEQEND"))
                        {
                            skipEntity();       // skip to end of SEQ end.

                            Point3f[]   points = new Point3f[verts.size()];

                            for (int i = 0; i != points.length; i++)
                            {
                                DXFVertex   v = (DXFVertex)verts.get(i);

                                points[i] = v.point;
                            }

                            if (extrusion)
                            {
                                Point3f[]   vs = new Point3f[4];

                                for (int i = 0; i != points.length - 1; i++)
                                {
                                    DXFVertex   v1 = (DXFVertex)verts.get(i);
                                    DXFVertex   v2 = (DXFVertex)verts.get(i + 1);
                                    
                                    if (!v1.point.equals(v2.point))
                                    {
                                        addPolygon(a, vs, v1.point, v2.point, extLen, dir, shader);
                                    }
                                }

                                if ((flags[BIT_FLAG] & CLOSED) != 0)
                                {
                                    DXFVertex   v1 = (DXFVertex)verts.get(points.length - 1);
                                    DXFVertex   v2 = (DXFVertex)verts.get(0);

                                    if (!v1.point.equals(v2.point))
                                    {
                                        addPolygon(a, vs, v1.point, v2.point, extLen, dir, shader);
                                    }
                                }
                            }
                            else if ((flags[BIT_FLAG] & THREEDPOLYMESH) != 0)
                            {
                                Point3f[]   tri = new Point3f[3];

                                for (int i = 0; i != flags[M] - 1; i++)
                                {
                                    for (int j = 0; j != flags[N] - 1; j++)
                                    {
                                        Point3f v1 = points[i * flags[N] + j];
                                        Point3f v2 = points[i * flags[N] + j + 1];

                                        if (v1.equals(v2))
                                        {
                                            continue;
                                        }

                                        Point3f v3 = points[(i + 1) * flags[N] + j + 1];
                                        if (v2.equals(v3))
                                        {
                                            continue;
                                        }

                                        Point3f v4 = points[(i + 1) * flags[N] + j];

                                        if (v3.equals(v4))
                                        {
                                            continue;
                                        }
                                        
                                        Point3f av = new Point3f(
                                                (v1.x + v2.x + v3.x + v4.x) / 4.0f,
                                                (v1.y + v2.y + v3.y + v4.y) / 4.0f,
                                                (v1.z + v2.z + v3.z + v4.z) / 4.0f);

                                        /*
                                         * at the moment we don't handle patches
                                         * properly so we break the area up
                                         * into triangles.
                                         */
                                        tri[0] = v1;
                                        tri[1] = v2;
                                        tri[2] = av;

                                        addTriangle(a, shader, tri);

                                        tri[0] = v2;
                                        tri[1] = v3;
                                        tri[2] = av;

                                        addTriangle(a, shader, tri);

                                        tri[0] = v3;
                                        tri[1] = v4;
                                        tri[2] = av;

                                        addTriangle(a, shader, tri);

                                        tri[0] = v4;
                                        tri[1] = v1;
                                        tri[2] = av;

                                        addTriangle(a, shader, tri);
                                    }
                                }
                            }
                            else if ((flags[BIT_FLAG] & POLYFACEMESH) != 0)
                            {
                                Point3f[]   tri = new Point3f[3];

                                for (int i = 0; i != points.length; i++)
                                {
                                    DXFVertex   v = (DXFVertex)verts.get(i);

                                    //
                                    // deal with "invisible" edges
                                    //
                                    if (v.flags[1] < 0)
                                    {
                                        v.flags[1] = -v.flags[1];
                                    }
                                    if (v.flags[2] < 0)
                                    {
                                        v.flags[2] = -v.flags[2];
                                    }
                                    if (v.flags[3] < 0)
                                    {
                                        v.flags[3] = -v.flags[3];
                                    }

                                    //
                                    // ignore lines and points
                                    //
                                    if (v.flags[1] == 0 || v.flags[2] == 0
                                        || v.flags[3] == 0)
                                    {
                                        continue;
                                    }

                                    tri[0] = points[v.flags[1] - 1];
                                    tri[1] = points[v.flags[2] - 1];
                                    tri[2] = points[v.flags[3] - 1];

                                    if (tri[0].equals(tri[1])
                                        || tri[1].equals(tri[2])
                                        || tri[2].equals(tri[0]))
                                    {
                                        continue;
                                    }

                                    Polygon p = new Polygon(tri);

                                    p.setShader(shader);

                                    a.add(p);
                                }
                            }
                            else if ((flags[BIT_FLAG] & CLOSED) != 0)
                            {
                                Polygon p = new Polygon(points);

                                p.setShader(shader);

                                a.add(p);
                            }

                            return;
                        }
                    }
                }

                line = in.readLine();

                switch (code)
                {
                case 8:         // layer
                    shader = (Shader)shaderTable.get(line);
                    break;
                case 6:
                case 62:
                    // color
                    break;
                case 5:
                case 67:
                    // handle
                    break;
                case 40:
                case 41:
                    break;
                case 66:
                    // entflag
                    break;
                case 10:
                    break;
                case 20:
                    break;
                case 30:
                    break;
                case 39:
                    extrusion = true;
                    extLen = Float.parseFloat(line);
                    break;
                case 70:
                case 71:
                case 72:
                case 73:
                case 74:
                case 75:
                    flags[code - 70] = Integer.parseInt(line.trim());
                    break;
                case 210:                 // EXTRU_X
                    extrusion = true;
                    dir.x = Float.parseFloat(line);
                    break;
                case 220:                 // EXTRU_Y
                    extrusion = true;
                    dir.y = Float.parseFloat(line);
                    break;
                case 230:                 // EXTRU_Z
                    extrusion = true;
                    dir.z = Float.parseFloat(line);
                    break;
                default:
                    System.out.println("unknown code encountered in polyline: " + code);
                }
            }

            throw new IOException("malformed polyline");
        }

        /**
         * read in a single 3d face
         */
        void read3DFace(
            Aggregate   a)
            throws IOException
        {
            Shader      shader = null;
            ArrayList   verts = new ArrayList();
            String      line;
            float       x, y, z;
            Point3f     last = null;
            int[]       flags = new int[NUM_FLAGS];

            x = y = z = 0;

            while ((line = in.readLine()) != null)
            {
                int code = Integer.parseInt(line.trim());
                if (code == 0)
                {
                    try
                    {
                        Polygon p = new Polygon(
                            (Point3f[])verts.toArray(new Point3f[verts.size()]));
                        p.setShader(shader);

                        a.add(p);
                    }
                    catch (IllegalArgumentException e)
                    {
                        // a dud - ignore
                    }
                    return;
                }

                line = in.readLine();

                switch (code)
                {
                case 8:         // layer
                    shader = (Shader)shaderTable.get(line);
                    break;
                case 62:
                    // color
                    break;
                case 5:
                case 67:
                    // handle
                    break;
                case 10:
                case 11:
                case 12:
                case 13:
                case 14:
                case 15:
                case 16:
                    x = Float.parseFloat(line);
                    break;
                case 20:
                case 21:
                case 22:
                case 23:
                case 24:
                case 25:
                case 26:
                    y = Float.parseFloat(line);
                    break;
                case 30:
                case 31:
                case 32:
                case 33:
                case 34:
                case 35:
                case 36:
                    z = Float.parseFloat(line);
                    //
                    // if it's a triangle the last point gets repeated...
                    //
                    if (last != null)
                    {
                        if (last.x != x || last.y != y || last.z != z)
                        {
                            last = new Point3f(x, y, z);
                            verts.add(last);
                        }
                    }
                    else
                    {
                        last = new Point3f(x, y, z);
                        verts.add(last);
                    }
                    break;
                case 70:
                case 71:
                case 72:
                case 73:
                case 74:
                case 75:
                    flags[code - 70] = Integer.parseInt(line.trim());
                    break;
                default:
                    System.out.println("unknown code encountered in 3D Face: " + code);
                }
            }

            throw new IOException("malformed 3D face");
        }

        /**
         *    skip the dxf entity, just go forward to look for the next
         * marker.
         */
        void skipEntity()
            throws IOException
        {
            int         c, i;
            String      line;

            while ((line = in.readLine()) != null)
            {
                if (Integer.parseInt(line.trim()) == 0)
                {
                    return;
                }

                line = in.readLine();
            }
        }

        void readEntities(
            Aggregate   a)
            throws IOException
        {
            String  line = in.readLine();
            int     code = Integer.parseInt(line.trim());

            while ((line = in.readLine()) != null)
            {
                if (line.equals("3DFACE"))
                {
                    read3DFace(a);
                }
                else if (line.equals("CIRCLE"))
                {
                    readCircle(a);
                }
                else if (line.equals("POLYLINE"))
                {
                    readPolyline(a);
                }
                else if (line.equals("LINE"))
                {
                    readLine(a);
                }
                else if (line.equals("SOLID"))
                {
                    readSolid(a);
                }
                else if (line.equals("INSERT"))
                {
                    readInsert(a);
                }
                else if (line.equals("EOF"))
                {
                    return;
                }
                else
                {
                    skipEntity();
                }
            }
        }

        void readObjects(
            Aggregate   a)
            throws IOException
        {
            boolean     entitiesBlockFound = false;
            String      line;

            while ((line = in.readLine()) != null)
            {
                if (line.equals("BLOCKS"))
                {
                    readBlocks();
                }
                else if (line.equals("ENTITIES"))
                {
                    entitiesBlockFound = true;
                    readEntities(a);
                }
            }

            if (!entitiesBlockFound)
            {
                throw new IOException("no entity block found!");
            }
        }
    }

    public DXFFile(
        String  geometry)
        throws IOException
    {
        this.geometry = geometry;

        addGeometry(geometry);
    }

    private DXFFile()
    {
    }

    public static Aggregate getInstance(
        RMLReader   in)
        throws IOException
    {
        DXFFile f = new DXFFile();

        f.readObject(in);

        f.addGeometry(f.geometry);

        return f;
    }

    protected void processObject(
        Object      obj)
    {
        super.processObject(obj);

        if (obj instanceof Shader)
        {
            Shader  s = (Shader)obj;

            shaderTable.put(s.getName(), s);
        }
    }

    protected void processParameters(
        ParametersMap   params)
    {
        super.processParameters(params);

        this.geometry = (String)params.get("geometry");
    }

    private void addGeometry(
        String    name)
        throws IOException
    {
        //
        // read in the previously unknown geometry
        //
        BufferedReader      in = new BufferedReader(new FileReader(name));
        DXFFileReader       r = new DXFFileReader(in, shaderTable);

        r.readObjects(this);
    }
}
