package org.autochthonous.graphics.rt.objects;

import java.io.*;

import javax.vecmath.*;

import org.autochthonous.math.*;
import org.autochthonous.math.eqn.*;

import org.autochthonous.graphics.rt.*;
import org.autochthonous.graphics.rt.util.*;
import org.autochthonous.graphics.rt.tools.io.*;

/**
 * Rendering for an arbitrary algebraic surface.
 * <p>
 * Sample RML:
 * <code>
 * <pre>
 * &lt;Algebraic color="0.1372, 0.557, 0.1372"
 *     equation="4(x^{2} + y^{2} + z^{2} - 13)^{3} + 27(3x^{2} + y^{2} - 4z^{2} - 12)^{2} = 0"/&gt;
 * </pre>
 * </code>
 */
public class Algebraic
    extends Renderable
    implements NormalCalculator
{
    static final double TOL = 0.0001;

    static EquationParser   parser = null;

    int                 order;
    double[]            roots;
    float[]             dists;

    Equation            e, dx, dy, dz;
    Term[]              eTerms;

    SturmSolver         solver = new SturmSolver();

    TextureMapper       stMapper;

    double              tolerance = TOL;

    /**
     * polynomial cache
     */
    RayPoly[]   xPolys;
    RayPoly[]   yPolys;
    RayPoly[]   zPolys;
    RayPoly[]   dxPolys;
    RayPoly[]   dyPolys;
    RayPoly[]   dzPolys;
    RayPoly     subTotal;
    RayPoly     total;

    protected class RayPoly
    {
        private int         order;
        private double[]    coefs, scratch;

        /**
         * order in this case is the maximum order this structure will need
         * to accomodate.
         */
        public RayPoly(
            int ord)
        {
            this.coefs = new double[ord + 1];
            this.scratch = new double[ord + 1];
            this.order = 0;
        }

        /**
         * raise x to the integer power e
         */
        private double power(
            double	x,
            int	    e)
        {
            double	a, b;

            switch (e) {
            case 0:
                return(1.0);
            case 1:
                return(x);
            case 2:
                return(x * x);
            case 3:
                return(x * x * x);
            case 4:
                b = x * x;
                return(b * b);
            case 5:
                b = x * x;
                return(b * b * x);
            case 6:
                b = x * x;
                return(b * b * b);
            default:
                b = x;
                a = 1;

                do
                {
                    if ((e & 1) != 0)
                    {
                        a *= b;
                    }

                    e /= 2;
                    b *= b;
                }
                while (e != 0);
            }

            return a;
        }

        private void addPow(
            double  o,
            double  d,
            int     pow)
        {
            double  o2, d2;

            switch (pow)
            {
            case 1:
                coefs[0] += o;
                coefs[1] += d;
                break;
            case 2:
                coefs[0] += o * o;
                coefs[1] += 2 * d * o;
                coefs[2] += d * d;
                break;
            case 3:
                coefs[0] += o * o;
                coefs[1] += 3 * d * o * o;
                coefs[2] += 3 * d * d * o;
                coefs[3] += d * d * d;
                break;
            case 4:
                o2 = o * o;
                d2 = d * d;
                coefs[0] += o2 * o2;
                coefs[1] += 4 * d * o2 * o;
                coefs[2] += 6 * d2 * o2;
                coefs[3] += 4 * d2 * d * o;
                coefs[4] += d2 * d2;
                break;
            case 5:
                o2 = o * o;
                d2 = d * d;
                coefs[0] += o * o2 * o2;
                coefs[1] += 5 * d * o2 * o2;
                coefs[2] += 10 * d2 * o2 * o;
                coefs[3] += 10 * d2 * d * o2;
                coefs[4] += 5 * d2 * d2 * o;
                coefs[5] += d * d2 * d2;
                break;
            default:
                for (int i = 0; i <= pow; i++)
                {
                    double  fact = power(o, pow - i) * power(d, i);
                    int     x = 2;

                    for (int j = pow; j > pow - i; j--)
                    {
                        fact *= j;
                        if (i > 2 && x <= i)
                        {
                            fact /= x;
                            x++;
                        }
                    }

                    while (x <= i)
                    {
                        fact /= x;
                        x++;
                    }

                    coefs[i] += fact;
                }
            }

            this.order = (pow > this.order) ? pow : this.order;
        }

        private void subPow(
            double  o,
            double  d,
            int     pow)
        {
            double  o2, d2;

            switch (pow)
            {
            case 1:
                coefs[0] -= o;
                coefs[1] -= d;
                break;
            case 2:
                coefs[0] -= o * o;
                coefs[1] -= 2 * d * o;
                coefs[2] -= d * d;
                break;
            case 3:
                coefs[0] -= o * o;
                coefs[1] -= 3 * d * o * o;
                coefs[2] -= 3 * d * d * o;
                coefs[3] -= d * d * d;
                break;
            case 4:
                o2 = o * o;
                d2 = d * d;
                coefs[0] -= o2 * o2;
                coefs[1] -= 4 * d * o2 * o;
                coefs[2] -= 6 * d2 * o2;
                coefs[3] -= 4 * d2 * d * o;
                coefs[4] -= d2 * d2;
                break;
            case 5:
                o2 = o * o;
                d2 = d * d;
                coefs[0] -= o * o2 * o2;
                coefs[1] -= 5 * d * o2 * o2;
                coefs[2] -= 10 * d2 * o2 * o;
                coefs[3] -= 10 * d2 * d * o2;
                coefs[4] -= 5 * d2 * d2 * o;
                coefs[5] -= d * d2 * d2;
                break;
            default:
                for (int i = 0; i <= pow; i++)
                {
                    double  fact = power(o, pow - i) * power(d, i);
                    int     x = 2;

                    for (int j = pow; j > pow - i; j--)
                    {
                        fact *= j;
                        if (i > 2 && x <= i)
                        {
                            fact /= x;
                            x++;
                        }
                    }

                    while (x <= i)
                    {
                        fact /= x;
                        x++;
                    }

                    coefs[i] -= fact;
                }
            }

            this.order = (pow > this.order) ? pow : this.order;
        }

        public int order()
        {
            return order;
        }

        public void addXpow(
            Ray r,
            int pow)
        {
            addPow(r.org.x, r.dir.x, pow);
        }

        public void addYpow(
            Ray r,
            int pow)
        {
            addPow(r.org.y, r.dir.y, pow);
        }

        public void addZpow(
            Ray r,
            int pow)
        {
            addPow(r.org.z, r.dir.z, pow);
        }

        public void subtractXpow(
            Ray r,
            int pow)
        {
            subPow(r.org.x, r.dir.x, pow);
        }

        public void subtractYpow(
            Ray r,
            int pow)
        {
            subPow(r.org.y, r.dir.y, pow);
        }

        public void subtractZpow(
            Ray r,
            int pow)
        {
            subPow(r.org.z, r.dir.z, pow);
        }

        public void zero()
        {
            for (int i = 0; i < coefs.length; i++)
            {
                coefs[i] = 0;
            }

            order = 0;
        }

        public void set(
            double v)
        {
            for (int i = 1; i < coefs.length; i++)
            {
                coefs[i] = 0;
            }

            coefs[0] = v;
            order = 0;
        }

        public void set(
            RayPoly p)
        {
            for (int i = p.order + 1; i < coefs.length; i++)
            {
                coefs[i] = 0;
            }

            for (int i = 0; i <= p.order; i++)
            {
                coefs[i] = p.coefs[i];
            }

            this.order = p.order;
        }

        public void add(
            double  v)
        {
            coefs[0] += v;
        }

        public void add(
            RayPoly  p)
        {
            for (int i = 0; i <= p.order; i++)
            {
                coefs[i] += p.coefs[i];
            }

            this.order = (p.order > this.order) ? p.order : this.order;
        }

        public void subtract(
            double  v)
        {
            coefs[0] -= v;
        }

        public void subtract(
            RayPoly  p)
        {
            for (int i = 0; i <= p.order; i++)
            {
                coefs[i] -= p.coefs[i];
            }
        }

        public void multiply(
            double  v)
        {
            for (int i = 0; i <= order; i++)
            {
                coefs[i] *= v;
            }
        }

        public void multiply(
            RayPoly  p)
        {
            for (int i = 0; i < scratch.length; i++)
            {
                scratch[i] = 0;
            }

            for (int i = 0; i <= order; i++)
            {
                double  a = coefs[i];

                for (int j = 0; j <= p.order; j++)
                {
                    scratch[i + j] += a * p.coefs[j];
                }
            }

            this.order += p.order;

            System.arraycopy(scratch, 0, coefs, 0, scratch.length);
        }

        public void square()
        {
            for (int i = 0; i < scratch.length; i++)
            {
                scratch[i] = 0;
            }

            for (int i = 0; i <= order; i++)
            {
                double  a = coefs[i];

                scratch[i + i] += a * a;

                for (int j = i + 1; j <= order; j++)
                {
                    scratch[i + j] += 2 * a * coefs[j];
                }
            }

            this.order += this.order;

            System.arraycopy(scratch, 0, coefs, 0, scratch.length);
        }

        public double[] getCoefficients()
        {
            return coefs;
        }
    }

    /**
     * base constructor for an algebraic surface.
     * <p>
     * The equation format follows the same rules as TeX, for example:
     * <pre>
     *  x^{2} + y^{2} + z^{2} = 1
     * </pre>
     * describes a unit sphere.
     */
    public Algebraic(
        String  eqn)
    {
        init(eqn);
    }

    protected Algebraic(
        int order)
    {
        this.order = order;
        this.roots = new double[order];
        this.dists = new float[order];
    }
        
    private Algebraic()
    {
    }

    public static Algebraic getInstance(
        RMLReader   in)
        throws IOException
    {
        Algebraic  a = new Algebraic();

        a.readObject(in);

        return a;
    }

    protected void processParameters(
        ParametersMap    params)
    {
        super.processParameters(params);

        if (params.get("equation") != null)
        {
            init(params.getString("equation", "x^{2} + y^{2} + z^{2} = 1"));
        }
    }

    private void addPolys(
        Term[]      terms,
        RayPoly[]   xPolys,
        RayPoly[]   yPolys,
        RayPoly[]   zPolys)
    {
        for (int i = 0; i != terms.length; i++)
        {
            Term    t = terms[i];

            if (t.xPow() > 0 && xPolys[t.xPow()] == null)
            {
                xPolys[t.xPow()] = new RayPoly(t.xPow());
            }

            if (t.yPow() > 0 && yPolys[t.yPow()] == null)
            {
                yPolys[t.yPow()] = new RayPoly(t.yPow());
            }

            if (t.zPow() > 0 && zPolys[t.zPow()] == null)
            {
                zPolys[t.zPow()] = new RayPoly(t.zPow());
            }
        }
    }

    private void init(
        String  eqn)
    {
        if (parser == null)
        {
            parser = new EquationParser(new CharArrayReader(eqn.toCharArray()));
        }
        else
        {
            EquationParser.ReInit(new CharArrayReader(eqn.toCharArray()));
        }

        try
        {
            this.e = EquationParser.equation();
        }
        catch (ParseException e)
        {
            throw new IllegalArgumentException("error parsing equation: " + eqn);
        }

        this.dx = e.dX();
        this.dy = e.dY();
        this.dz = e.dZ();

        this.eTerms = e.getTerms();

        this.order = 0;

        //
        // find out what order of a polynomial we will create
        //
        for (int i = 0; i != eTerms.length; i++)
        {
            Term    t = eTerms[i];
            int     ord = t.xPow() + t.yPow() + t.zPow();

            if (ord > order)
            {
                order = ord;
            }
        }

        this.roots = new double[order];
        this.dists = new float[order];

        this.total = new RayPoly(order);
        this.subTotal = new RayPoly(order);
        this.xPolys = new RayPoly[order + 1];
        this.yPolys = new RayPoly[order + 1];
        this.zPolys = new RayPoly[order + 1];

        addPolys(eTerms, xPolys, yPolys, zPolys);
    }

    private int algebraicIntersect(
        Ray         r,
        float[]     dists,
        boolean     findAllRoots)
    {
        int        np, numroots, atmin, atinf;
        int        its;
        float      t, min;
        Point3f    org;
        Vector3f   dir;

        org = r.org;
        dir = r.dir;

        RayPoly     p = getPolynomial(r);
        double[]    coef = p.getCoefficients();

        //
        // if all the coefficients are > 0.0 we can't have any
        // positive roots!
        //
        boolean negFound = false;

        for (int i = 0; i != coef.length; i++)
        {
            if (coef[i] < 0)
            {
                negFound = true;
                break;
            }
        }

        if (!negFound)
        {
            return 0;
        }

        solver.init(coef);

        int changes = solver.changesAtX(tolerance) - solver.changesAtX(Double.POSITIVE_INFINITY);

        if (changes == 0)
        {
            return 0;
        }

        if (!findAllRoots)
        {
            double  max = 2.0;

            while ((solver.changesAtX(tolerance) - solver.changesAtX(max)) == 0)
            {
                max *= 2;
                if (max > r.tMax)
                {
                    break;
                }
            }

            if (!solver.findRoot(tolerance, max, roots))
            {
                return 0;
            }

            dists[0] = (float)(roots[0]);

            return 1;
        }
        else
        {
            double  max = 2.0;

            while ((solver.changesAtX(tolerance) - solver.changesAtX(max)) != changes)
            {
                max *= 2;
                if (max > r.tMax)
                {
                    break;
                }
            }

            int     nRoots = solver.findRoots(tolerance, max, roots);

            if (nRoots == 0)
            {
                return 0;
            }

            for (int i = 0; i != nRoots; i++)
            {
                dists[i] = (float)(roots[i]);
            }

            return nRoots;
        }
    }

    public boolean isShadowing(
        Ray     r)
    {
        Ray     nr = this.createObjectSpaceRay(r);

        tolerance = TOL;

        /*
        if (tolerance < r.tMin)
        {
            tolerance = r.tMin;
        }
        */

        if (algebraicIntersect(nr, dists, false) == 0)
        {
            nr.dispose();

            return false;
        }

        nr.dispose();

        return true;
    }

    public HitList intersections(
        Ray     r)
    {
        Ray     nr = this.createObjectSpaceRay(r);

        tolerance = TOL;

        /*
        if (tolerance < r.tMin)
        {
            tolerance = r.tMin;
        }
        */

        int nRoots = algebraicIntersect(nr, dists, true);

        if (nRoots == 0)
        {
            nr.dispose();

            return null;
        }

        HitList list = null;

        for (int i = 0; i != nRoots; i++)
        {
            if (dists[i] > tolerance)
            {
                if (list == null)
                {
                    list = HitList.getInstance();
                }

                list.add(r, this, dists[i], this, this, stMapper, this.getShader());
            }
        }

        nr.dispose();

        return list;
    }

    public boolean isIntersected(
        Ray     r)
    {
        Ray     nr = this.createObjectSpaceRay(r);

        tolerance = TOL;

        if (tolerance < r.tMin)
        {
            tolerance = r.tMin;
        }

        if (algebraicIntersect(nr, dists, false) == 0)
        {
            nr.dispose();

            return false;
        }

        r.setHit(this, dists[0], this, this, stMapper, this.getShader());

        nr.dispose();

        return true;
    }

    protected RayPoly getPolynomial(
        Ray r)
    {
        for (int i = 0; i != xPolys.length; i++)
        {
            if (xPolys[i] != null)
            {
                xPolys[i].zero();
                xPolys[i].addXpow(r, i);
            }
        }

        for (int i = 0; i != yPolys.length; i++)
        {
            if (yPolys[i] != null)
            {
                yPolys[i].zero();
                yPolys[i].addYpow(r, i);
            }
        }

        for (int i = 0; i != zPolys.length; i++)
        {
            if (zPolys[i] != null)
            {
                zPolys[i].zero();
                zPolys[i].addZpow(r, i);
            }
        }

        total.zero();

        for (int i = 0; i != eTerms.length; i++)
        {
            Term    t = eTerms[i];

            subTotal.set(t.coefficient());

            if (t.xPow() > 0)
            {
                subTotal.multiply(xPolys[t.xPow()]);
            }

            if (t.yPow() > 0)
            {
                subTotal.multiply(yPolys[t.yPow()]);
            }

            if (t.zPow() > 0)
            {
                subTotal.multiply(zPolys[t.zPow()]);
            }

            total.add(subTotal);
        }

        return total;
    }

    public void getN(Hit hit, Vector3f n)
    {
        Point3f     loc = hit.oP();

        n.set((float)dx.eval(loc.x, loc.y, loc.z),
                (float)dy.eval(loc.x, loc.y, loc.z),
                (float)dz.eval(loc.x, loc.y, loc.z));
    }
}
