package org.autochthonous.graphics.rt.objects;

import java.io.*;

import javax.vecmath.*;

import org.autochthonous.graphics.rt.*;
import org.autochthonous.graphics.rt.util.*;
import org.autochthonous.graphics.rt.tools.io.*;

/**
 * A bounding box object - whatever it contains will be clipped to the
 * box dimensions.
 * <p>
 * Example:
 * <code>
 * <pre>
 *      &lt;BoundingBox&gt;
 *          &lt;Algebraic color="0.1372, 0.557, 0.1372"
 *              equation="0.5(x^{5} + x^{4}) - y^{2} - z^{2} = 0"/&gt;
 *      &lt;/BoundingBox&gt;
 * </pre>
 * </code>
 * <p>
 * Document Type Definition:
 * <code>
 * <pre>
 *     &lt;!ENTITY %Renderable  " Aggregate | Box | Cone | Cylinder | IndexedMesh | Polygon | Sphere | SuperQuadric | Torus | TriangularStrip | Algebraic "&gt;
 *
 *     &lt;!ELEMENT BoundingBox %Renderable; &gt;
 * </pre>
 * </code>
 */
public class BoundingBox
    extends Bounded
{
    BBox                    bBox = null;
    Renderable              obj;
    float[]                 vals = new float[2];

    public BoundingBox(
        BBox        bBox,
        Renderable  obj)
    {
        this.bBox = bBox;
        this.obj = obj;
    }

    public static BoundingBox getInstance(
        RMLReader   in)
        throws IOException
    {
        Object          o;
        Renderable      obj = null;
        ParametersMap   params = in.parameters();

        Point3f p1 = params.getPoint("vertex1", "0, 0, 0");
        Point3f p2 = params.getPoint("vertex2", "1, 1, 1");

        BBox    bb = new BBox(p1.x, p1.y, p1.z, p2.x, p2.y, p2.z);

        while ((o = in.readObject()) != null)
        {
            obj = (Renderable)o;
        }

        return new BoundingBox(bb, obj);
    }

    public BBox bBox()
    {
        return bBox;
    }

    public void transform(
         Matrix4f   trans)
    {
        super.transform(trans);

        obj.transform(trans);
    }

    public void setShader(
        Shader  s)
    {
        super.setShader(s);

        obj.setShader(s);
    }

    public boolean isIntersected(
        Ray    r)
    {
        Ray     nr = this.createObjectSpaceRay(r);
        boolean hit = false;

        if (bBox.isIntersecting(nr.org, nr.dir, vals))
        {
            float   tMin = r.tMin;
            float   tMax = r.tMax;

            r.tMin = vals[0];
            r.tMax = vals[1];

            hit = obj.isIntersected(r);

            r.tMin = tMin;
            r.tMax = tMax;
        }

        nr.dispose();

        return hit;
    }

    public HitList intersections(
        Ray    r)
    {
        Ray     nr = this.createObjectSpaceRay(r);
        HitList hits = null;

        if (bBox.isIntersecting(nr.org, nr.dir, vals))
        {
            float   tMin = r.tMin;
            float   tMax = r.tMax;

            r.tMin = vals[0];
            r.tMax = vals[1];

            hits = obj.intersections(r);

            r.tMin = tMin;
            r.tMax = tMax;
        }

        nr.dispose();

        return hits;
    }

    public boolean isShadowing(
        Ray    r)
    {
        Ray     nr = this.createObjectSpaceRay(r);
        boolean ret = false;

        if (bBox.isIntersecting(nr.org, nr.dir, vals))
        {
            float   tMin = r.tMin;
            float   tMax = r.tMax;

            r.tMin = vals[0];
            r.tMax = vals[1];

            ret = obj.isShadowing(r);

            r.tMin = tMin;
            r.tMax = tMax;
        }

        nr.dispose();

        return ret;
    }
}
