package org.autochthonous.graphics.rt.viewers;

import org.autochthonous.graphics.rt.*;

/**
 * Class for the changing camera with linear interpolation
 * <p>
 * Sample RML for a changing camera is:
 * <code>
 * <pre>
 *      &lt;ChangingCamera&gt;
 *          &lt;Camera&gt;
 *              &lt;Perspective fov="45.0"/&gt;
 *              &lt;LookAt eye="0, 0, 5" reference="0, 0, 0" up="0, 1, 0"/&gt;
 *          &lt;/Camera&gt;
 *          &lt;Camera&gt;
 *              &lt;Perspective fov="45.0"/&gt;
 *              &lt;LookAt eye="0, 0, 6" reference="0, 0, 0" up="0, 1, 0"/&gt;
 *          &lt;/Camera&gt;
 *      &lt;/ChangingCamera&gt;
 * </pre>
 * </code>
 * <p>
 */
public class LinearChangingCamera
    extends ChangingCamera
{
    public LinearChangingCamera(
        Camera[]  choices)
    {
        super(choices);
    }

    /**
     * @param sampleNo number of this sample.
     * @param time time in the frame of this sample (0.0 to 1.0)
     * @param x x position on screen
     * @param y y position on screen
     */
    public Ray generateRay(
        int         sampleNo,
        float       time,
        float       x,
        float       y)
    {
        int         index = (int)(time * (choices.length - 1));

        if (index == choices.length - 1)
        {
            index = choices.length - 2;
        }

        Camera      a = choices[index];
        Camera      b = choices[index + 1];

        Ray         r = Ray.getInstance();
        Ray         rA = Ray.getInstance();
        Ray         rB = Ray.getInstance();

        rA = a.generateRay(sampleNo, time, x, y);
        rB = b.generateRay(sampleNo, time, x, y);

        float   fact = (time * (choices.length - 1)) - index;

        r.org.set((rA.org.x + (rB.org.x - rA.org.x) * fact),
                    (rA.org.y + (rB.org.y - rA.org.y) * fact),
                    (rA.org.z + (rB.org.z - rA.org.z) * fact)); 

        r.dir.set((rA.dir.x + (rB.dir.x - rA.dir.x) * fact),
                    (rA.dir.y + (rB.dir.y - rA.dir.y) * fact),
                    (rA.dir.z + (rB.dir.z - rA.dir.z) * fact)); 

        r.dir.normalize();

        rA.dispose();
        rB.dispose();

        return r;
    }
}
