package org.autochthonous.graphics.tools;

import java.io.*;
import java.util.*;

import javax.vecmath.*;

import org.autochthonous.graphics.rt.*;
import org.autochthonous.graphics.rt.objects.*;
import org.autochthonous.graphics.rt.shaders.CookTorranceShader;

/**
 * a dump utility to show properties in the old 3ds files.
 */
public class ThreeDStudioDump
{
    /**
     * A 3-d studio style surface shader
     */
    private static class ThreeDStudioShader
        extends CookTorranceShader
    {
        protected Color3f       diffuse = new Color3f();
        protected Color3f       specular = new Color3f();
        protected Color3f       reflectance = new Color3f();

        protected float         roughness = 0;
        protected float         reflectionRoughness = 0;

        protected boolean       reflectingSurface = false;

        protected int           maxLevel = 6;

        public ThreeDStudioShader(
            World   world)
        {
            super(world);
        }

        /**
         * set the name for this shader
         */
        public void setName(
            String    name)
        {
            this.name = name;
        }

        /**
         * set the ambient components for this object
         */
        public void setAmbient(
            Color3f c)
        {
            ambient.set(c.x, c.y, c.z);
        }

        /**
         * set the diffuse components for this object
         */
        public void setDiffuse(
            Color3f c)
        {
            diffuse.set(c.x, c.y, c.z);
        }

        /**
         * set the specular components for this object
         */
        public void setSpecular(
            Color3f c)
        {
            specular.set(c.x, c.y, c.z);
        }

        /**
         * set the roughness of the surface, the smaller we are the more
         * refined the specular highlights.
         */
        public void setShininess(
            double    shininess)
        {
            if (shininess > 0.01)
            {
                this.roughness = (float)(1.0 - shininess);
            }
        }

        /**
         * return whether, or not, the surface is specular.
         */
        public boolean isSpecular()
        {
            return (roughness != 0);
        }

        /**
         * set the reflective component for this surface.
         */
        public void setReflectance(
            double    r,
            double    g,
            double    b)
        {
            reflectance.set((float)r, (float)g, (float)b);
            if ((r != 0) || (g != 0) || (b != 0))
            {
                reflectingSurface = true;
            }
        }

        /**
         * set the reflective component for this surface.
         */
        public void setReflectance(
            Color3f c)
        {
            setReflectance(c.x, c.y, c.z);
        }

        /**
         * return whether, or not, the surface is reflecting.
         */
        public boolean isReflecting()
        {
            return reflectingSurface;
        }

        public void setReflectionRoughness(
            double    reflectionRoughness)
        {
            this.reflectionRoughness = (float)reflectionRoughness;
        }

        /**
         * the shade routine. If you want more than the below, this is
         * the method to override.
         */
        public void shade(
            Hit        hit,
            Pixel      p)
        {
            p.set(0, 0, 0, 1);

            p.scaledAdd(hit.color(), ambientContribution(world));

            LightDataList   contribs = hit.visibleLights(world);

            p.scaledAdd(hit.color(), diffuse, diffuseContribution(contribs, hit));

            if (isSpecular())
            {
                p.add(specularContribution(contribs, hit, roughness));
            }

            //
            // with secondary rays there is always the chance we
            // can recurse into infinity, so we have to set a hard limit
            // to stop at.
            //
            if (hit.level() >= maxLevel)
            {
                return;
            }

            if (reflectionRoughness != 0)
            {
                hit.perturbN(reflectionRoughness);
            }

            //
            // fire off a reflection ray (if necessary).
            //
            if (isReflecting())
            {
                Ray      ir = hit.reflectedRay();
                Pixel    pix = new Pixel();

                ir.trace(world, pix);

                p.add(pix);

                ir.dispose();
            }

            p.scale(alpha);
        }
    }

    static private class ThreeDStudioReader
    {
        InputStream in;
        HashMap     shaderTable = new HashMap();
        World       world;
        HashMap     ignoreFlags = new HashMap();

        private static final int UNKNOWN_1              = 0x0002;
        private static final int MAIN_CHUNK             = 0x4d4d;
        private static final int MASTER_SCALE_CHUNK     = 0x0100;
        private static final int OBJMESH_CHUNK          = 0x3d3d;
        private static final int UNKNOWN_2              = 0x3d3e;
        private static final int MATERIAL_CHUNK         = 0xafff;
        private static final int MATERIAL_NAME_CHUNK    = 0xa000;
        private static final int MATERIAL_MAPNAME_CHUNK = 0xa300;
        private static final int AMBIENT_COLOR_CHUNK    = 0xa010;
        private static final int DIFFUSE_COLOR_CHUNK    = 0xa020;
        private static final int SPECULAR_COLOR_CHUNK   = 0xa030;
        private static final int SHININESS_CHUNK        = 0xa040;
        private static final int SHININESS2_CHUNK       = 0xa041;
        private static final int TRANSPARENCY_CHUNK     = 0xa050;
        private static final int RGB_FLOAT_CHUNK        = 0x0010;
        private static final int RGB_BYTE_CHUNK         = 0x0011;
        private static final int RGB_BYTE_LIN_CHUNK     = 0x0012;
        private static final int PERCENT_INT_CHUNK      = 0x0030;
        private static final int PERCENT_FLOAT_CHUNK    = 0x0031;
        private static final int OBJECT_BLOCK_CHUNK     = 0x4000;
        private static final int TRIMESH_CHUNK          = 0x4100;
        private static final int VERTEX_LIST_CHUNK      = 0x4110;
        private static final int FACE_LIST_CHUNK        = 0x4120;
        private static final int MESH_MAT_CHUNK         = 0x4130;
        private static final int TEX_VERTEX_CHUNK       = 0x4140;
        private static final int SMOOTH_CHUNK           = 0x4150;
        private static final int TRANSFORM_CHUNK        = 0x4160;
        private static final int MESH_COLOR_CHUNK       = 0x4165;
        private static final int DIRECT_LIGHT_CHUNK     = 0x4600;
        private static final int CAMERA_CHUNK           = 0x4700;
        private static final int KF_CHUNK               = 0xb000;
        private static final int KFHDR_CHUNK            = 0xb00a;
        private static final int VIEWPORT_LAYOUT_CHUNK  = 0x7001;
        private static final int KF_SEG_CHUNK           = 0xb008;
        private static final int KF_CURTIME_CHUNK       = 0xb009;
        private static final int OBJECT_NODE_TAG_CHUNK  = 0xb002;
        private static final int NODEHDR_CHUNK          = 0xb010;
        private static final int DUMMY_CHUNK            = 0xffff;

        ThreeDStudioReader(
            InputStream in)
        {
            this.in = in;
        }

        int read16()
            throws IOException
        {
            return (in.read() | (in.read() << 8));
        }

        long read32()
            throws IOException
        {
            return (in.read() | (in.read() << 8) | (in.read() << 16) | ((long)in.read() << 24));
        }
            
        float readFloat()
            throws IOException
        {
            return Float.intBitsToFloat(in.read() | (in.read() << 8) | (in.read() << 16) | (in.read() << 24));
        }

        String readName()
            throws IOException
        {
            ByteArrayOutputStream   bOut = new ByteArrayOutputStream();
            int                     c;

            while ((c = in.read()) > 0)
            {
                bOut.write(c);
            }

            return new String(bOut.toByteArray());
        }

        float readPercentage()
            throws IOException
        {
            int     id = read16();
            long    size = read32();

            switch (id)
            {
            case PERCENT_INT_CHUNK:
                return (read16() / 100.0f);
            case PERCENT_FLOAT_CHUNK:
                return (readFloat() / 100.0f);
            default:
                skipChunk(size);
            }

            return 0;
        }

        Color3f readColor()
            throws IOException
        {
            int     id = read16();
            long    size = read32();

            switch (id)
            {
            case RGB_FLOAT_CHUNK:
                return new Color3f(readFloat(), readFloat(), readFloat());
            case RGB_BYTE_CHUNK:
            case RGB_BYTE_LIN_CHUNK:
                return new Color3f(in.read() / 255.0f, in.read() / 255.0f, in.read() / 255.0f);
            default:
                skipChunk(size);
            }

            return null;
        }

        Shader readMaterial(
            long    length)
            throws IOException
        {
            ThreeDStudioShader  shader = new ThreeDStudioShader(world);

            length -= 6;

            do
            {
                int     id = read16();

                if (id < 0)
                {
                    break;
                }

                long    size = read32();

                length -= size;

                switch(id)
                {
                case MATERIAL_NAME_CHUNK:
                    shader.setName(readName());
                    break;
                case AMBIENT_COLOR_CHUNK:
                    shader.setAmbient(readColor());
                    break;
                case DIFFUSE_COLOR_CHUNK:
                    shader.setDiffuse(readColor());
                    break;
                case SPECULAR_COLOR_CHUNK:
                    shader.setSpecular(readColor());
                    break;
                case SHININESS_CHUNK:
                    float shininess1 = readPercentage();
                    shader.setShininess(shininess1);
                    break;
                case SHININESS2_CHUNK:
                    float shininess2 = readPercentage();
                    break;
                case TRANSPARENCY_CHUNK:
                    float transparency = readPercentage();
                    break;
                case DUMMY_CHUNK:
                    skipChunk(size);
                    break;
                default:
                    skipChunk(size);
                    break;
                }
            }
            while (length > 0);

            if (shaderTable.get(shader.getName()) == null)
            {
                shaderTable.put(shader.getName(), shader);
            }

            return shader;
        }

        void readLight()
            throws IOException
        {
            float    x, y, z;
            
            x = readFloat();
            y = readFloat();
            z = readFloat();

            // 
            // Now read the color chunk
            //
            int     id = read16();
            long    size = read32();

            readColor();
        }

        void readCamera()
            throws IOException
        {
            float eyex = readFloat();
            float eyey = readFloat();
            float eyez = readFloat();

            float focx = readFloat();
            float focy = readFloat();
            float focz = readFloat();

            float twist = readFloat();    /* Not sure on this one */
            float n = readFloat();            /* Not sure on this one */
        }

        Object readVerticies(
            ArrayList   verts)
            throws IOException
        {
            int         nVerts, i;
            float       x, y, z;

            nVerts = read16();

            for (i = 0; i < nVerts; i++)
            {
                x = readFloat();
                y = readFloat();
                z = readFloat();

                verts.add(new Point3f(x, y, z));
            }

            return null;
        }

        Object readFaces(
            long        length,
            Aggregate   aggregate,
            ArrayList   polys,
            ArrayList   verts,
            ArrayList   shaders)
            throws IOException
        {
            int     nPoly, i, n, a, b, c, d;

            nPoly = read16();

            n = 0;
            for (i = 0; i < nPoly; i++)
            {
                a = read16();
                b = read16();
                c = read16();
                d = read16();    // some dodgey flag value 

                short[] v = new short[3];

                v[0] = (short)a;
                v[1] = (short)b;
                v[2] = (short)c;

                polys.add(v);
                shaders.add(null);    // expand shaderIndexes array
            }

            length -= nPoly * 8 + 2;

            do
            {
                int id = read16();
                if (id < 0)
                {
                    return null;
                }

                long size = read32();

                length -= size;

                switch(id)
                {
                case MESH_MAT_CHUNK:
                    readMeshMaterials(size, shaders);
                    break;
                case SMOOTH_CHUNK:
                    // readSmoothGroup(size);
                    skipChunk(size);
                    break;
                case OBJECT_BLOCK_CHUNK:
                    readObjectBlock(size, aggregate);
                    break;
                default:
                    skipChunk(size);
                }
            }
            while (length > 0);

            return null;
        }

        void readTransform()
            throws IOException
        {
            float[][] m = new float[4][];

            m[0] = new float[4];
            m[1] = new float[4];
            m[2] = new float[4];
            m[3] = new float[4];

            m[0][0] = readFloat();
            m[0][1] = readFloat();
            m[0][2] = readFloat();

            m[1][0] = readFloat();
            m[1][1] = readFloat();
            m[1][2] = readFloat();

            m[2][0] = readFloat();
            m[2][1] = readFloat();
            m[2][2] = readFloat();

            m[3][0] = readFloat();
            m[3][1] = readFloat();
            m[3][2] = readFloat();
        }

        void readMeshMaterials(
            long        size,
            ArrayList   shaders)
            throws IOException
        {
            Shader  shader = (Shader)shaderTable.get(readName());

            int nFaces = read16();

            for (int i = 0; i < nFaces; i++)
            {
                int f = read16();
                shaders.set(f, shader);
            }
        }

        void readTextureVertices()
            throws IOException
        {
            int nVerts = read16();

            for (int i = 0; i < nVerts; i++)
            {
                readFloat();
                readFloat();
            }
        }

        void readSmoothGroup()
            throws IOException
        {
        }

        void readTriMesh(
            long        length,
            Aggregate   aggregate)
            throws IOException
        {
            length -= 6;

            ArrayList   polys = new ArrayList();
            ArrayList   verts = new ArrayList();
            ArrayList   shaders = new ArrayList();

            do
            {
                int     id = read16();

                if (id < 0)
                {
                    break;
                }

                long    size = read32();

                length -= size;

                switch (id)
                {
                case FACE_LIST_CHUNK:
                    readFaces(size, aggregate, polys, verts, shaders);
                    break;
                case VERTEX_LIST_CHUNK:
                    readVerticies(verts);
                    break;
                case TRANSFORM_CHUNK:
                    readTransform();
                    break;
                case MESH_COLOR_CHUNK:
                    int c = in.read();
                    break;
                case TEX_VERTEX_CHUNK:
                    readTextureVertices();
                    break;
                case DUMMY_CHUNK:
                    skipChunk(size);
                    break;
                default:
                    skipChunk(size);
                }
            }
            while (length > 0);

            short[][]   ps = (short[][])polys.toArray(new short[polys.size()][]);
            Point3f[]   vs = (Point3f[])verts.toArray(new Point3f[verts.size()]);
            IndexedMesh mesh = new IndexedMesh(vs, ps);

            if (ignoreFlags.get("materials") == null)
            {
                HashMap     indexes = new HashMap();
                short       index = 0;
                ArrayList   shaderArray = new ArrayList();

                Iterator    it = shaderTable.keySet().iterator();
                while (it.hasNext())
                {
                    Object  shader = shaderTable.get(it.next());

                    indexes.put(shader, new Short(index++));
                    shaderArray.add(shader);
                }

                short[]     ss = new short[shaders.size()];

                for (int i = 0; i != ss.length; i++)
                {
                    ss[i] = ((Short)indexes.get(shaders.get(i))).shortValue();
                }

                mesh.setShader(
                 (Shader[])shaderArray.toArray(new Shader[shaderArray.size()]), ss);
            }

            aggregate.add(mesh);
        }

        Object readObjectBlock(
            long        length,
            Aggregate   aggregate)
            throws IOException
        {
            length -= 6;

            if (length == 0)
            {
                return null;
            }

            String  n = readName();

            length -= n.length() + 1;

            do
            {
                int id = read16();
                if (id < 0)
                {
                    return null;
                }

                long size = read32();

                length -= size;

                switch(id)
                {
                case TRIMESH_CHUNK:
                    readTriMesh(size, aggregate);
                    break;
                default:
                    skipChunk(size);
                }
            }
            while (length > 0);

            return null;
        }

        Object readObjMeshChunk(
            long        length,
            Aggregate   aggregate)
            throws IOException
        {
            length -= 6;

            if (length == 0)
            {
                return null;
            }

            do
            {
                int id = read16();
                if (id < 0)
                {
                    break;
                }

                long size = read32();

                length -= size;

                switch(id)
                {
                case UNKNOWN_2:
                    skipChunk(size);
                    break;
                case MASTER_SCALE_CHUNK:
                    float mscale = readFloat();
                    break;
                case MATERIAL_CHUNK:
                    readMaterial(size);
                    break;
                case OBJECT_BLOCK_CHUNK:
                    readObjectBlock(size, aggregate);
                    break;
                default:
                    skipChunk(size);
                }
            }
            while (length > 0);

            return null;
        }

        Object readChunk(
            long        length,
            Aggregate   aggregate)
            throws IOException
        {
            length -= 6;

            if (length == 0)
            {
                return null;
            }

            do
            {
                int id = read16();
                if (id < 0)
                {
                    return null;
                }

                long size = read32();

                length -= size;
                switch(id)
                {
                case UNKNOWN_1:
                    skipChunk(size);
                    break;
                case OBJMESH_CHUNK:
                    readObjMeshChunk(size, aggregate);
                    break;
                case KF_CHUNK:
                    readChunk(size, aggregate);
                    break;
                default:
                    System.out.println("unknown chunk ignored (" + Integer.toString(id, 16) + ")");
                    skipChunk(size);
                }
            }
            while (length > 0);

            return new Integer(1);
        }

        void readChunk()
            throws IOException
        {
            int id = read16();
            if (id < 0)
            {
                return;
            }

            Aggregate   aggregate = new Aggregate();

            long size = read32();
            switch(id)
            {
            case MAIN_CHUNK:
                readChunk(size, aggregate);
                break;
            default:
                System.out.println("unknown chunk ignored in main (" + Integer.toString(id, 16) + ")");
                skipChunk(size);
            }

            System.out.println("Bounding Box");
            System.out.println(aggregate.bBox());
            System.out.println("Materials");
            Iterator    it = shaderTable.keySet().iterator();
            while (it.hasNext())
            {
                Shader  s = (Shader)shaderTable.get(it.next());

                System.out.println("material: '" + s.getName() + "'");
            }
        }

        void skipChunk(
            long    chunkSize)
            throws IOException
        {
            in.skip(chunkSize - 6);
        }
    }

    public static void main(
        String[]    args)
    {
        if (args.length != 1)
        {
            System.err.println("usage: 3dsdump file_name");
            System.exit(1);
        }

        try
        {
            FileInputStream     in = new FileInputStream(args[0]);
            ThreeDStudioReader  r = new ThreeDStudioReader(in);

            r.readChunk();
        }
        catch (EOFException e)
        {
            System.err.println("3dsdump: can't open file " + args[0]);
            System.exit(1);
        }
        catch (IOException e)
        {
            System.err.println("3dsdump: problem reading file " + args[0]
                        + " - " + e.toString());
            System.exit(1);
        }
    }
}
