package org.autochthonous.math.eqn;

import java.util.ArrayList;

/**
 * general class for manipulating equations in x, y, and z, symbolicly.
 * <p>
 * To use this class effectively you should assume the result of the
 * equation you buid up should be zero.
 */
public class Equation
{
    ArrayList   terms;
    int         order;

    /**
     * create an equation that equates to a constant
     */
    public Equation(
        double  v)
    {
        terms = new ArrayList();
        terms.add(new Term(v));
        order = 0;
    }

    /**
     * create an equation that equates to x, y, or z.
     */
    public Equation(
        char    axis)
    {
        terms = new ArrayList();
        terms.add(new Term(axis));
        order = 1;
    }

    /**
     * combine common terms together.
     */
    private void prune()
    {
        int ord = 0;

        for (int i = 0; i != terms.size(); i++)
        {
            Term    t1 = (Term)terms.get(i);

            int j = i + 1;

            if (t1.order() > ord)
            {
                ord = t1.order();
            }

            while (j < terms.size())
            {
                Term    t2 = (Term)terms.get(j);

                if (t1.xPow() == t2.xPow()
                        && t1.yPow() == t2.yPow() && t1.zPow() == t2.zPow())
                {
                    t1 = new Term(t1.coefficient() + t2.coefficient(), t1.xPow(), t1.yPow(), t1.zPow());
                    terms.set(i, t1);
                    terms.remove(j);
                }
                else
                {
                    j++;
                }
            }

            this.order = ord;
        }
    }

    private Equation(
        ArrayList   terms)
    {
        this.terms = terms;

        prune();
    }

    public int order()
    {
        return this.order;
    }

    public Equation add(
        Equation e)
    {
        ArrayList   nTerms = new ArrayList();

        for (int i = 0; i != this.terms.size(); i++)
        {
            nTerms.add((Term)this.terms.get(i));
        }

        for (int i = 0; i != e.terms.size(); i++)
        {
            nTerms.add(e.terms.get(i));
        }

        return new Equation(nTerms);
    }

    public Equation subtract(
        Equation e)
    {
        ArrayList   nTerms = new ArrayList();

        for (int i = 0; i != this.terms.size(); i++)
        {
            nTerms.add((Term)this.terms.get(i));
        }

        for (int i = 0; i != e.terms.size(); i++)
        {
            Term    t = (Term)e.terms.get(i);

            nTerms.add(new Term(t.coefficient() * -1, t.xPow(), t.yPow(), t.zPow()));
        }

        return new Equation(nTerms);
    }

    public Equation multiply(
        double v)
    {
        ArrayList   nTerms = new ArrayList();

        for (int i = 0; i != terms.size(); i++)
        {
            Term    t = (Term)terms.get(i);

            nTerms.add(new Term(t.coefficient() * v, t.xPow(), t.yPow(), t.zPow()));
        }

        return new Equation(nTerms);
    }

    public Equation multiply(
        Equation e)
    {
        ArrayList   nTerms = new ArrayList();

        for (int i = 0; i != terms.size(); i++)
        {
            for (int j = 0; j != e.terms.size(); j++)
            {
                Term    t1 = (Term)terms.get(i);
                Term    t2 = (Term)e.terms.get(j);

                nTerms.add(new Term(t1.coefficient() * t2.coefficient(), t1.xPow() + t2.xPow(), t1.yPow() + t2.yPow(), t1.zPow() + t2.zPow()));
            }
        }

        return new Equation(nTerms);
    }

    public Equation divide(
        double v)
    {
        ArrayList   nTerms = new ArrayList();

        for (int i = 0; i != terms.size(); i++)
        {
            Term    t = (Term)terms.get(i);

            nTerms.add(new Term(t.coefficient() / v, t.xPow(), t.yPow(), t.zPow()));
        }

        return new Equation(nTerms);
    }

    public Equation power(
        int pow)
    {
        if (pow == 0)
        {
            return new Equation(1.0);
        }

        if (pow == 1)
        {
            return this;
        }

        Equation    res = this;

        for (int i = 1; i != pow; i++)
        {
            res = res.multiply(this);
        }

        return res;
    }

    /**
     * return the partial derivative of this equation in X.
     */
    public Equation dX()
    {
        ArrayList   nTerms = new ArrayList();

        for (int i = 0; i != terms.size(); i++)
        {
            Term    t = (Term)terms.get(i);

            if (t.xPow() >= 1)
            {
                nTerms.add(new Term(t.coefficient() * t.xPow(), t.xPow() - 1, t.yPow(), t.zPow()));
            }
        }

        if (nTerms.size() == 0)
        {
            return new Equation(0.0);
        }

        return new Equation(nTerms);
    }

    /**
     * return the partial derivative of this equation in Y.
     */
    public Equation dY()
    {
        ArrayList   nTerms = new ArrayList();

        for (int i = 0; i != terms.size(); i++)
        {
            Term    t = (Term)terms.get(i);

            if (t.yPow() >= 1)
            {
                nTerms.add(new Term(t.coefficient() * t.yPow(), t.xPow(), t.yPow() - 1, t.zPow()));
            }
        }

        if (nTerms.size() == 0)
        {
            return new Equation(0.0);
        }

        return new Equation(nTerms);
    }

    /**
     * return the partial derivative of this equation in Z.
     */
    public Equation dZ()
    {
        ArrayList   nTerms = new ArrayList();

        for (int i = 0; i != terms.size(); i++)
        {
            Term    t = (Term)terms.get(i);

            if (t.zPow() >= 1)
            {
                nTerms.add(new Term(t.coefficient() * t.zPow(), t.xPow(), t.yPow(), t.zPow() - 1));
            }
        }

        if (nTerms.size() == 0)
        {
            return new Equation(0.0);
        }

        return new Equation(nTerms);
    }

    /**
     * return the terms for the equation
     */
    public Term[] getTerms()
    {
        return (Term[])terms.toArray(new Term[terms.size()]);
    }

    /**
     * evaluate the equation for a given x, y, and z
     */
    public double eval(
        double  x,
        double  y,
        double  z)
    {
        double  total = 0;

        for (int i = 0; i != terms.size(); i++)
        {
            Term    t = (Term)terms.get(i);

            total += t.coefficient() * Math.pow(x, t.xPow()) * Math.pow(y, t.yPow()) * Math.pow(z, t.zPow());
        }

        return total;
    }

    public String toString()
    {
        StringBuffer    b = new StringBuffer();
        boolean         first = true;

        for (int i = 0; i != terms.size(); i++)
        {
            Term    t = (Term)terms.get(i);

            if (first)
            {
                first = false;
            }
            else
            {
                if (t.coefficient() > 0.0)
                {
                    b.append("+");
                }
            }

            b.append(t.coefficient());

            if (t.xPow() >= 1)
            {
                b.append("x");
                if (t.xPow() > 1)
                {
                    b.append("^{");
                    b.append(t.xPow());
                    b.append("}");
                }
            }

            if (t.yPow() >= 1)
            {
                b.append("y");
                if (t.yPow() > 1)
                {
                    b.append("^{");
                    b.append(t.yPow());
                    b.append("}");
                }
            }

            if (t.zPow() >= 1)
            {
                b.append("z");
                if (t.zPow() > 1)
                {
                    b.append("^{");
                    b.append(t.zPow());
                    b.append("}");
                }
            }
        }

        return b.toString();
    }
}
